@@ -15,21 +15,23 @@ from io import open
 from time import time
 from tqdm import tqdm
 from math import pow, tanh
-from .util import Print, VERBOSITY, Preproc as Pp
+from .util import is_a_collection, Print, VERBOSITY, Preproc as Pp
 
 # python 2 and 3 compatibility
 from functools import reduce
 from six.moves import xrange
 
-__version__ = "0.5.6"
+__version__ = "0.6.1"
 
 ENCODING = "utf-8"
 
 PARA_DELTR = "\n"
 SENT_DELTR = r"\."
 WORD_DELTR = r"\s"
+WORD_REGEX = r"\w+(?:'\w+)?"
 
 STR_UNKNOWN, STR_MOST_PROBABLE = "unknown", "most-probable"
+STR_OTHERS_CATEGORY = "[others]"
 STR_UNKNOWN_CATEGORY = "[unknown]"
 IDX_UNKNOWN_CATEGORY = -1
 STR_UNKNOWN_WORD = ''
@@ -94,6 +96,8 @@ class SS3:
     __p__ = 1
     __a__ = .0
 
+    __multilabel__ = False
+
     __l_update__ = None
     __s_update__ = None
     __p_update__ = None
@@ -111,6 +115,7 @@ class SS3:
     __parag_delimiter__ = PARA_DELTR
     __sent_delimiter__ = SENT_DELTR
     __word_delimiter__ = WORD_DELTR
+    __word_regex__ = WORD_REGEX
 
     def __init__(
         self, s=None, l=None, p=None, a=None,
@@ -138,13 +143,19 @@ class SS3:
         :param sn_m: method used to compute the sanction (sn) function, options
                      are: "vanilla" and "xai" (default: "xai")
         :type sn_m: str
+        :raises: ValueError
         """
         self.__name__ = (name or self.__name__).lower()
 
-        self.__s__ = s or self.__s__
-        self.__l__ = l or self.__l__
-        self.__p__ = p or self.__p__
-        self.__a__ = a or self.__a__
+        self.__s__ = self.__s__ if s is None else s
+        self.__l__ = self.__l__ if l is None else l
+        self.__p__ = self.__p__ if p is None else p
+        self.__a__ = self.__a__ if a is None else a
+
+        try:
+            float(self.__s__ + self.__l__ + self.__p__ + self.__a__)
+        except BaseException:
+            raise ValueError("hyperparameter values must be numbers")
 
         self.__categories_index__ = {}
         self.__categories__ = []
@@ -361,120 +372,128 @@ class SS3:
         cv[:] = [(v if v > self.__a__ else 0) for v in cv]
         return cv
 
-    def __classify_sentence__(self, sent, prep, json=False):
+    def __classify_sentence__(self, sent, prep, json=False, prep_func=None):
         """Classify the given sentence."""
         classify_trans = self.__classify_ngram__
-        cats = xrange(len(self.__categories__))
+        categories = self.__categories__
+        cats = xrange(len(categories))
         word_index = self.get_word_index
         word_delimiter = self.__word_delimiter__
+        word_regex = self.__word_regex__
+
         if not json:
-            if prep:
-                sent = Pp.clean_and_ready(sent)
+            if prep or prep_func is not None:
+                prep_func = prep_func or Pp.clean_and_ready
+                sent = prep_func(sent)
             sent_words = [
                 (w, w)
-                for w in re.split(word_delimiter, sent)
+                for w in re_split_keep(word_regex, sent)
                 if w
             ]
         else:
-            if prep:
+            if prep or prep_func is not None:
                 sent_words = [
-                    (w, Pp.clean_and_ready(w, dots=False))
-                    for w in re_split_keep(word_delimiter, sent)
+                    (w, Pp.clean_and_ready(w, dots=False) if prep_func is None else prep_func(w))
+                    for w in re_split_keep(word_regex, sent)
                     if w
                 ]
             else:
                 sent_words = [
                     (w, w)
-                    for w in re_split_keep(word_delimiter, sent)
+                    for w in re_split_keep(word_regex, sent)
                     if w
                 ]
 
         if not sent_words:
             sent_words = [(u'.', u'.')]
 
-        flat_sent = []
-        flat_raw_sent = []
-        for raw_seq, seq in sent_words:
-            words = re.split(word_delimiter, seq)
-            for iw in xrange(len(words)):
-                word = words[iw]
-
-                if json and re.match(self.__word_delimiter__, raw_seq):
-                    if len(flat_raw_sent):
-                        flat_raw_sent[-1] += raw_seq
-                        continue
-
-                wordi = word_index(word)
-
-                if iw == len(words) - 1:
-                    word_iend = len(raw_seq)
-                else:
-                    if not word.startswith("nnbrr"):
-                        try:
-                            word_iend = re.search(word, raw_seq, re.I).end()
-                        except AttributeError:
-                            word_iend = len(word)
-                    else:
-                        word_iend = re.search(r"\d+", raw_seq).end()
-
-                flat_sent.append(wordi)
-                flat_raw_sent.append(raw_seq[:word_iend])
-                raw_seq = raw_seq[word_iend:]
-
-        sent = []
-        fs_len = len(flat_sent)
+        sent_iwords = [word_index(w) for _, w in sent_words]
+        sent_len = len(sent_iwords)
+        sent_parsed = []
         wcur = 0
-        while wcur < fs_len:
+        while wcur < sent_len:
+            cats_ngrams_cv = [[0] for icat in cats]
+            cats_ngrams_offset = [[0] for icat in cats]
+            cats_ngrams_iword = [[-1] for icat in cats]
+            cats_max_cv = [.0 for icat in cats]
 
-            cats_ngrams = [[] for icat in cats]
-            cats_max = [.0 for icat in cats]
             for icat in cats:
                 woffset = 0
-                word = flat_sent[wcur]
-                word_info = self.__categories__[icat][VOCAB]
-                while word in word_info:
-                    cats_ngrams[icat].append(word_info[word][CV])
-                    word_info = word_info[word][NEXT]
-                    woffset += 1
-                    if wcur + woffset >= fs_len:
-                        break
-                    word = flat_sent[wcur + woffset]
+                word_raw = sent_words[wcur + woffset][0]
+                wordi = sent_iwords[wcur + woffset]
+                word_info = categories[icat][VOCAB]
 
-                cats_max[icat] = (max(cats_ngrams[icat])
-                                  if cats_ngrams[icat] else .0)
+                if wordi in word_info:
+                    cats_ngrams_cv[icat][0] = word_info[wordi][CV]
+                    word_info = word_info[wordi][NEXT]
+                cats_ngrams_iword[icat][0] = wordi
+                cats_ngrams_offset[icat][0] = woffset
 
-            max_gv = max(cats_max)
+                # if it is a learned word (not unknown and seen for this category),
+                # then try to recognize learned n-grams too
+                if wordi != IDX_UNKNOWN_WORD and wordi in categories[icat][VOCAB]:
+                    # while word or word delimiter (e.g. space)
+                    while wordi != IDX_UNKNOWN_WORD or re.match(word_delimiter, word_raw):
+                        woffset += 1
+                        if wcur + woffset >= sent_len:
+                            break
+
+                        word_raw = sent_words[wcur + woffset][0]
+                        wordi = sent_iwords[wcur + woffset]
+
+                        # if word is a word:
+                        if wordi != IDX_UNKNOWN_WORD:
+                            # if this word belongs to this category
+                            if wordi in word_info:
+                                cats_ngrams_cv[icat].append(word_info[wordi][CV])
+                                cats_ngrams_iword[icat].append(wordi)
+                                cats_ngrams_offset[icat].append(woffset)
+                                word_info = word_info[wordi][NEXT]
+                            else:
+                                break
+
+                    cats_max_cv[icat] = (max(cats_ngrams_cv[icat])
+                                         if cats_ngrams_cv[icat] else .0)
+
+            max_gv = max(cats_max_cv)
+            use_ngram = True
             if (max_gv > self.__a__):
-                cat_max_gv = cats_max.index(max_gv)
-                ngram_len = cats_ngrams[cat_max_gv].index(max_gv) + 1
+                icat_max_gv = cats_max_cv.index(max_gv)
+                ngram_max_gv = cats_ngrams_cv[icat_max_gv].index(max_gv)
+                offset_max_gv = cats_ngrams_offset[icat_max_gv][ngram_max_gv] + 1
 
                 max_gv_sum_1_grams = max([
                     sum([
-                        self.__categories__[ic][VOCAB][w][CV]
-                        if w in self.__categories__[ic][VOCAB]
-                        else 0
-                        for w
-                        in flat_sent[wcur:wcur + ngram_len]
+                        (categories[ic][VOCAB][wi][CV]
+                         if wi in categories[ic][VOCAB]
+                         else 0)
+                        for wi
+                        in cats_ngrams_iword[ic]
                     ])
                     for ic in cats
                 ])
 
                 if (max_gv_sum_1_grams > max_gv):
-                    ngram_len = 1
+                    use_ngram = False
             else:
-                ngram_len = 1
+                use_ngram = False
 
-            sent.append(
+            if not use_ngram:
+                offset_max_gv = 1
+                icat_max_gv = 0
+                ngram_max_gv = 0
+
+            sent_parsed.append(
                 (
-                    flat_raw_sent[wcur:wcur + ngram_len],
-                    flat_sent[wcur:wcur + ngram_len]
+                    u"".join([raw_word for raw_word, _ in sent_words[wcur:wcur + offset_max_gv]]),
+                    cats_ngrams_iword[icat_max_gv][:ngram_max_gv + 1]
                 )
             )
-            wcur += ngram_len
+            wcur += offset_max_gv
 
         get_word = self.get_word
         if not json:
-            words_cvs = [classify_trans(seq) for _, seq in sent]
+            words_cvs = [classify_trans(seq) for _, seq in sent_parsed]
             if words_cvs:
                 return self.summary_op_ngrams(words_cvs)
             return self.__zero_cv__
@@ -484,12 +503,12 @@ class SS3:
             info = [
                 {
                     "token": u"→".join(map(get_word, sequence)),
-                    "lexeme": u"".join(raw_sequence),
+                    "lexeme": raw_sequence,
                     "cv": classify_trans(sequence),
                     "lv": [local_value(sequence, ic) for ic in cats],
                     "fr": [get_tip(sequence, ic)[FR] for ic in cats]
                 }
-                for raw_sequence, sequence in sent
+                for raw_sequence, sequence in sent_parsed
             ]
             return {
                 "words": info,
@@ -497,11 +516,11 @@ class SS3:
                 "wmv": reduce(vmax, [v["cv"] for v in info])  # word max value
             }
 
-    def __classify_paragraph__(self, parag, prep, json=False):
+    def __classify_paragraph__(self, parag, prep, json=False, prep_func=None):
         """Classify the given paragraph."""
         if not json:
             sents_cvs = [
-                self.__classify_sentence__(sent, prep=prep)
+                self.__classify_sentence__(sent, prep=prep, prep_func=prep_func)
                 for sent in re.split(self.__sent_delimiter__, parag)
                 if sent
             ]
@@ -510,7 +529,7 @@ class SS3:
             return self.__zero_cv__
         else:
             info = [
-                self.__classify_sentence__(sent, prep=prep, json=True)
+                self.__classify_sentence__(sent, prep=prep, prep_func=prep_func, json=True)
                 for sent in re_split_keep(self.__sent_delimiter__, parag)
                 if sent
             ]
@@ -543,8 +562,6 @@ class SS3:
 
         If category name doesn't exist, creates a new one.
         """
-        if isinstance(name, str):
-            name = name.lower()
         try:
             return self.__categories_index__[name]
         except KeyError:
@@ -616,6 +633,17 @@ class SS3:
         )
         return sorted(vocab_icat, key=lambda k: -k[-1])
 
+    def __get_def_cat__(self, def_cat):
+        """Given the `def_cat` argument, get the default category value."""
+        if def_cat is not None and (def_cat not in [STR_MOST_PROBABLE, STR_UNKNOWN] and
+                                    self.get_category_index(def_cat) == IDX_UNKNOWN_CATEGORY):
+            raise ValueError(
+                "the default category must be 'most-probable', 'unknown', or a category name"
+                " (current value is '%s')." % str(def_cat)
+            )
+        def_cat = None if def_cat == STR_UNKNOWN else def_cat
+        return self.get_most_probable_category() if def_cat == STR_MOST_PROBABLE else def_cat
+
     def __get_next_iwords__(self, sent, icat):
         """Return the list of possible following words' indexes."""
         if not self.get_category_name(icat):
@@ -784,6 +812,9 @@ class SS3:
             if def_cat == IDX_UNKNOWN_CATEGORY:
                 raise InvalidCategoryError
 
+        # does the special "[others]" category exist? (only used in multilabel classification)
+        __other_idx__ = self.get_category_index(STR_OTHERS_CATEGORY)
+
         if self.__update_needed__():
             self.update_values()
 
@@ -837,24 +868,29 @@ class SS3:
             else:
                 if pred_cv.sum() == 0:
                     if def_cat == IDX_UNKNOWN_CATEGORY:
-                        y_pred[doc_idx] = [STR_UNKNOWN_CATEGORY if labels else def_cat]
+                        y_pred[doc_idx] = []
                     else:
                         y_pred[doc_idx] = [self.get_category_name(def_cat) if labels else def_cat]
                 else:
-                    r = sorted(
-                        [
-                            (i, pred_cv[i])
-                            for i in range(pred_cv.size)
-                        ],
-                        key=lambda e: -e[1]
-                    )
+                    r = sorted([(i, pred_cv[i])
+                                for i in range(pred_cv.size)],
+                               key=lambda e: -e[1])
                     if labels:
-                        y_pred[doc_idx] = [
-                            self.get_category_name(cat_i)
-                            for cat_i, _ in r[:kmean_multilabel_size(r)]
-                        ]
+                        y_pred[doc_idx] = [self.get_category_name(cat_i)
+                                           for cat_i, _ in r[:kmean_multilabel_size(r)]]
                     else:
                         y_pred[doc_idx] = [cat_i for cat_i, _ in r[:kmean_multilabel_size(r)]]
+
+                # if the special "[others]" category exists
+                if __other_idx__ != IDX_UNKNOWN_CATEGORY:
+                    # if its among the predicted labels, remove (hide) it
+                    if labels:
+                        if STR_OTHERS_CATEGORY in y_pred[doc_idx]:
+                            y_pred[doc_idx].remove(STR_OTHERS_CATEGORY)
+                    else:
+                        if __other_idx__ in y_pred[doc_idx]:
+                            y_pred[doc_idx].remove(__other_idx__)
+
         return y_pred
 
     def summary_op_ngrams(self, cvs):
@@ -1159,7 +1195,7 @@ class SS3:
         """
         return self.__a__
 
-    def get_categories(self):
+    def get_categories(self, all=False):
         """
         Get the list of category names.
 
@@ -1169,6 +1205,7 @@ class SS3:
         return [
             self.get_category_name(ci)
             for ci in range(len(self.__categories__))
+            if all or self.get_category_name(ci) != STR_OTHERS_CATEGORY
         ]
 
     def get_most_probable_category(self):
@@ -1200,8 +1237,6 @@ class SS3:
         :rtype: int
         """
         try:
-            if isinstance(name, str):
-                name = name.lower()
             return self.__categories_index__[name]
         except KeyError:
             return IDX_UNKNOWN_CATEGORY
@@ -1362,7 +1397,8 @@ class SS3:
             "__index_to_word__": self.__index_to_word__,
             "__word_to_index__": self.__word_to_index__,
             "__cv_mode__": self.__cv_mode__,
-            "__sn_mode__": self.__sn_mode__
+            "__sn_mode__": self.__sn_mode__,
+            "__multilabel__": self.__multilabel__
         }
 
         try:
@@ -1417,21 +1453,22 @@ class SS3:
                 STR_MODEL_EXT
             ), "r", encoding=ENCODING
         )
-        json_file_format = json.loads(json_file.read(), object_hook=key_as_int)
+        jmodel = json.loads(json_file.read(), object_hook=key_as_int)
         json_file.close()
 
-        self.__max_fr__ = json_file_format["__max_fr__"]
-        self.__max_gv__ = json_file_format["__max_gv__"]
-        self.__l__ = json_file_format["__l__"]
-        self.__p__ = json_file_format["__p__"]
-        self.__s__ = json_file_format["__s__"]
-        self.__a__ = json_file_format["__a__"]
-        self.__categories__ = json_file_format["__categories__"]
-        self.__categories_index__ = json_file_format["__categories_index__"]
-        self.__index_to_word__ = json_file_format["__index_to_word__"]
-        self.__word_to_index__ = json_file_format["__word_to_index__"]
-        self.__cv_mode__ = json_file_format["__cv_mode__"]
-        self.__sn_mode__ = json_file_format["__sn_mode__"]
+        self.__max_fr__ = jmodel["__max_fr__"]
+        self.__max_gv__ = jmodel["__max_gv__"]
+        self.__l__ = jmodel["__l__"]
+        self.__p__ = jmodel["__p__"]
+        self.__s__ = jmodel["__s__"]
+        self.__a__ = jmodel["__a__"]
+        self.__categories__ = jmodel["__categories__"]
+        self.__categories_index__ = jmodel["__categories_index__"]
+        self.__index_to_word__ = jmodel["__index_to_word__"]
+        self.__word_to_index__ = jmodel["__word_to_index__"]
+        self.__cv_mode__ = jmodel["__cv_mode__"]
+        self.__sn_mode__ = jmodel["__sn_mode__"]
+        self.__multilabel__ = jmodel["__multilabel__"] if "__multilabel__" in jmodel else False
 
         self.__zero_cv__ = (0,) * len(self.__categories__)
         self.__s_update__ = self.__s__
@@ -1805,6 +1842,7 @@ class SS3:
         :raises: InvalidCategoryError, ValueError
         """
         r = self.classify(doc, json=True)
+        word_regex = self.__word_regex__
 
         if cat == 'auto':
             c_i = r["cvns"][0][0]
@@ -1859,7 +1897,7 @@ class SS3:
                             if words[w_i]["cv"][c_i] > min_cv:
                                 ww_left += min(ww_size, (len(words) - 1) - w_i)
 
-                            if re.search(r"[\w\d]+", words[w_i]["lexeme"]):
+                            if re.search(word_regex, words[w_i]["lexeme"]):
                                 ww_left -= 1
 
                             w_i += 1
@@ -1891,13 +1929,16 @@ class SS3:
                         ``2`` means 1-grams and 2-grams,
                         ``3``, 1-grams, 2-grams and 3-grams, and so on.
         :type n_grams: int
-        :param prep: enables input preprocessing (default: True)
+        :param prep: enables the default input preprocessing (default: True)
         :type prep: bool
         :param update: enables model auto-update after learning (default: True)
         :type update: bool
         """
         self.__cv_cache__ = None
 
+        if not doc or not cat:
+            return
+
         try:
             doc = doc.decode(ENCODING)
         except UnicodeEncodeError:  # for python 2 compatibility
@@ -1908,13 +1949,14 @@ class SS3:
         icat = self.__get_category__(cat)
         cat = self.__categories__[icat]
         word_to_index = self.__word_to_index__
+        word_regex = self.__word_regex__
 
         if prep:
             Print.info("preprocessing document...", offset=1)
             stime = time()
             doc = Pp.clean_and_ready(doc)
             Print.info("finished --time: %.1fs" % (time() - stime), offset=1)
-        doc = doc.replace("\n", "").split(" ")
+        doc = re.findall("%s|[^%s]+" % (word_regex, self.__word_delimiter__), doc)
 
         text_len = len(doc)
         Print.info(
@@ -1931,7 +1973,7 @@ class SS3:
         Print.info("learning...", offset=1)
         tips = []
         for word in doc:
-            if word and word != '.':
+            if re.match(word_regex, word):
                 self.__prun_counter__ += 1
                 # if word doesn't exist yet, then...
                 try:
@@ -1988,18 +2030,23 @@ class SS3:
         if update:
             self.update_values(force=True)
 
-    def classify(self, doc, prep=True, sort=True, json=False):
+    def classify(self, doc, prep=True, sort=True, json=False, prep_func=None):
         """
         Classify a given document.
 
         :param doc: the content of the document
         :type doc: str
-        :param prep: enables input preprocessing (default: True)
+        :param prep: enables the default input preprocessing (default: True)
         :type prep: bool
         :param sort: sort the classification result (from best to worst)
         :type sort: bool
         :param json: return a debugging version of the result in JSON format.
         :type json: bool
+        :param prep_func: the custom preprocessing function to be applied to
+                          the given document before classifying it.
+                          If not given, the default preprocessing function will
+                          be used (as long as ``prep=True``)
+        :type prep_func: function
         :returns: the document confidence vector if ``sort`` is False.
                   If ``sort`` is True, a list of pairs
                   (category index, confidence value) ordered by confidence value.
@@ -2022,7 +2069,7 @@ class SS3:
 
         if not json:
             paragraphs_cvs = [
-                self.__classify_paragraph__(parag, prep=prep)
+                self.__classify_paragraph__(parag, prep=prep, prep_func=prep_func)
                 for parag in re.split(self.__parag_delimiter__, doc)
                 if parag
             ]
@@ -2041,7 +2088,7 @@ class SS3:
             return cv
         else:
             info = [
-                self.__classify_paragraph__(parag, prep=prep, json=True)
+                self.__classify_paragraph__(parag, prep=prep, prep_func=prep_func, json=True)
                 for parag in re_split_keep(self.__parag_delimiter__, doc)
                 if parag
             ]
@@ -2082,7 +2129,7 @@ class SS3:
         :param labels: whether to return the category label or just the
                        category index (default: True)
         :type labels: bool
-        :param prep: enables input preprocessing (default: True)
+        :param prep: enables the default input preprocessing process (default: True)
         :type prep: bool
         :returns: the category label or the category index.
         :rtype: str or int
@@ -2104,7 +2151,7 @@ class SS3:
 
         return cat if labels else self.get_category_index(cat)
 
-    def classify_multilabel(self, doc, def_cat=STR_MOST_PROBABLE, labels=True, prep=True):
+    def classify_multilabel(self, doc, def_cat=STR_UNKNOWN, labels=True, prep=True):
         """
         Classify a given document returning multiple category labels.
 
@@ -2117,12 +2164,12 @@ class SS3:
         :param def_cat: default category to be assigned when SS3 is not
                         able to classify a document. Options are
                         "most-probable", "unknown" or a given category name.
-                        (default: "most-probable")
+                        (default: "unknown")
         :type def_cat: str
         :param labels: whether to return the category labels or just the
                        category indexes (default: True)
         :type labels: bool
-        :param prep: enables input preprocessing (default: True)
+        :param prep: enables the default input preprocessing (default: True)
         :type prep: bool
         :returns: the list of category labels (or indexes).
         :rtype: list (of str or int)
@@ -2132,22 +2179,31 @@ class SS3:
 
         if not r or not r[0][1]:
             if not def_cat or def_cat == STR_UNKNOWN:
-                cat = STR_UNKNOWN_CATEGORY
+                return []
             elif def_cat == STR_MOST_PROBABLE:
                 cat = self.get_most_probable_category()
             else:
                 if self.get_category_index(def_cat) == IDX_UNKNOWN_CATEGORY:
                     raise InvalidCategoryError
                 cat = def_cat
-            return [cat] if labels else [self.get_category_index(cat)]
-        else:
-            if labels:
-                return [
-                    self.get_category_name(cat_i)
-                    for cat_i, _ in r[:kmean_multilabel_size(r)]
-                ]
+            if cat != STR_OTHERS_CATEGORY:
+                return [cat] if labels else [self.get_category_index(cat)]
             else:
-                return [cat_i for cat_i, _ in r[:kmean_multilabel_size(r)]]
+                return []
+        else:
+            __other_idx__ = self.get_category_index(STR_OTHERS_CATEGORY)
+            if labels:
+                result = [self.get_category_name(cat_i)
+                          for cat_i, _ in r[:kmean_multilabel_size(r)]]
+                # removing "hidden" special category ("[other]")
+                if __other_idx__ != IDX_UNKNOWN_CATEGORY and STR_OTHERS_CATEGORY in result:
+                    result.remove(STR_OTHERS_CATEGORY)
+            else:
+                result = [cat_i for cat_i, _ in r[:kmean_multilabel_size(r)]]
+                # removing "hidden" special category ("[other]")
+                if __other_idx__ != IDX_UNKNOWN_CATEGORY and __other_idx__ in result:
+                    result.remove(__other_idx__)
+            return result
 
     def fit(self, x_train, y_train, n_grams=1, prep=True, leave_pbar=True):
         """
@@ -2156,47 +2212,77 @@ class SS3:
         :param x_train: the list of documents
         :type x_train: list (of str)
         :param y_train: the list of document labels
-        :type y_train: list (of str)
+        :type y_train: list of str for singlelabel classification;
+                        list of list of str for multilabel classification.
         :param n_grams:  indicates the maximum ``n``-grams to be learned
                         (e.g. a value of ``1`` means only 1-grams (words),
                         ``2`` means 1-grams and 2-grams,
                         ``3``, 1-grams, 2-grams and 3-grams, and so on.
         :type n_grams: int
-        :param prep: enables input preprocessing (default: True)
+        :param prep: enables the default input preprocessing (default: True)
         :type prep: bool
         :param leave_pbar: controls whether to leave the progress bar or
                            remove it after finishing.
         :type leave_pbar: bool
+        :raises: ValueError
         """
-        cats = sorted(list(set(y_train)))
         stime = time()
-
         x_train, y_train = list(x_train), list(y_train)
 
-        x_train = [
-            "".join([
-                x_train[i]
-                if x_train[i] and x_train[i][-1] == '\n'
-                else
-                x_train[i] + '\n'
-                for i in xrange(len(x_train))
-                if y_train[i] == cat
-            ])
-            for cat in cats
-        ]
-        y_train = list(cats)
+        if len(x_train) != len(y_train):
+            raise ValueError("`x_train` and `y_train` must have the same length")
+
+        if len(y_train) == 0:
+            raise ValueError("`x_train` and `y_train` are empty")
+
+        # if it's a multi-label classification problem
+        if is_a_collection(y_train[0]):
+            # flattening y_train
+            labels = [l for y in y_train for l in y]
+            self.__multilabel__ = True
+        else:
+            labels = y_train
+
+        cats = sorted(list(set(labels)))
+
+        # if it's a single-label classification problem
+        if not is_a_collection(y_train[0]):
+            x_train = [
+                "".join([
+                    x_train[i]
+                    if x_train[i] and x_train[i][-1] == '\n'
+                    else
+                    x_train[i] + '\n'
+                    for i in xrange(len(x_train))
+                    if y_train[i] == cat
+                ])
+                for cat in cats
+            ]
+            y_train = list(cats)
 
         Print.info("about to start training", offset=1)
         Print.verbosity_region_begin(VERBOSITY.NORMAL)
         progress_bar = tqdm(total=len(x_train), desc="Training",
                             leave=leave_pbar, disable=Print.is_quiet())
-        for i in range(len(x_train)):
-            progress_bar.set_description_str("Training on '%s'" % str(y_train[i]))
-            self.learn(
-                x_train[i], y_train[i],
-                n_grams=n_grams, prep=prep, update=False
-            )
-            progress_bar.update(1)
+
+        # if it's a multi-label classification problem
+        if is_a_collection(y_train[0]):
+            __others__ = [STR_OTHERS_CATEGORY]
+            for i in range(len(x_train)):
+                for label in (y_train[i] if y_train[i] else __others__):
+                    self.learn(
+                        x_train[i], label,
+                        n_grams=n_grams, prep=prep, update=False
+                    )
+                progress_bar.update(1)
+        else:
+            for i in range(len(x_train)):
+                progress_bar.set_description_str("Training on '%s'" % str(y_train[i]))
+                self.learn(
+                    x_train[i], y_train[i],
+                    n_grams=n_grams, prep=prep, update=False
+                )
+                progress_bar.update(1)
         progress_bar.close()
         self.__prune_tries__()
         Print.verbosity_region_end()
@@ -2209,7 +2295,7 @@ class SS3:
 
         :param x_test: the list of documents to be classified
         :type x_test: list (of str)
-        :param prep: enables input preprocessing (default: True)
+        :param prep: enables the default input preprocessing (default: True)
         :type prep: bool
         :param leave_pbar: controls whether to leave the progress bar after
                            finishing or remove it.
@@ -2233,7 +2319,7 @@ class SS3:
         ]
 
     def predict(
-        self, x_test, def_cat=STR_MOST_PROBABLE,
+        self, x_test, def_cat=None,
         labels=True, multilabel=False, prep=True, leave_pbar=True
     ):
         """
@@ -2244,6 +2330,8 @@ class SS3:
         :param def_cat: default category to be assigned when SS3 is not
                         able to classify a document. Options are
                         "most-probable", "unknown" or a given category name.
+                        (default: "most-probable", or "unknown" for
+                         multi-label classification)
         :type def_cat: str
         :param labels: whether to return the list of category names or just
                        category indexes
@@ -2251,8 +2339,10 @@ class SS3:
         :param multilabel: whether to perform multi-label classification or not.
                            if enabled, for each document returns a ``list`` of labels
                            instead of a single label (``str``).
+                           If the model was trained using multilabeled data, then this
+                           argument will be ignored and set to True.
         :type multilabel: bool
-        :param prep: enables input preprocessing (default: True)
+        :param prep: enables the default input preprocessing (default: True)
         :type prep: bool
         :param leave_pbar: controls whether to leave the progress bar or
                            remove it after finishing.
@@ -2265,11 +2355,17 @@ class SS3:
         if not self.__categories__:
             raise EmptyModelError
 
+        multilabel = multilabel or self.__multilabel__
+
+        if def_cat is None:
+            def_cat = STR_UNKNOWN if multilabel else STR_MOST_PROBABLE
+
         if not def_cat or def_cat == STR_UNKNOWN:
-            Print.info(
-                "default category was set to 'unknown' (its index will be -1)",
-                offset=1
-            )
+            if not multilabel:
+                Print.info(
+                    "default category was set to 'unknown' (its index will be -1)",
+                    offset=1
+                )
         else:
             if def_cat == STR_MOST_PROBABLE:
                 Print.info(
@@ -2313,10 +2409,11 @@ class SS3:
         [*] the gv function is defined in Section 3.2.2 of the original paper:
          https://arxiv.org/pdf/1905.08772.pdf
 
-        Example
-        >>> clf.cv("chicken", "food")
-        >>> clf.cv("roast chicken", "food")
-        >>> clf.cv("chicken", "sports")
+        Examples:
+
+            >>> clf.cv("chicken", "food")
+            >>> clf.cv("roast chicken", "food")
+            >>> clf.cv("chicken", "sports")
 
         :param ngram: the word or word n-gram
         :type ngram: str
@@ -2335,10 +2432,11 @@ class SS3:
         (gv function is defined in Section 3.2.2 of the original paper:
          https://arxiv.org/pdf/1905.08772.pdf)
 
-        Example
-        >>> clf.gv("chicken", "food")
-        >>> clf.gv("roast chicken", "food")
-        >>> clf.gv("chicken", "sports")
+        Examples:
+
+            >>> clf.gv("chicken", "food")
+            >>> clf.gv("roast chicken", "food")
+            >>> clf.gv("chicken", "sports")
 
         :param ngram: the word or word n-gram
         :type ngram: str
@@ -2357,10 +2455,11 @@ class SS3:
         (lv function is defined in Section 3.2.2 of the original paper:
          https://arxiv.org/pdf/1905.08772.pdf)
 
-        Example
-        >>> clf.lv("chicken", "food")
-        >>> clf.lv("roast chicken", "food")
-        >>> clf.lv("chicken", "sports")
+        Examples:
+
+            >>> clf.lv("chicken", "food")
+            >>> clf.lv("roast chicken", "food")
+            >>> clf.lv("chicken", "sports")
 
         :param ngram: the word or word n-gram
         :type ngram: str
@@ -2379,10 +2478,11 @@ class SS3:
         (sg function is defined in Section 3.2.2 of the original paper:
          https://arxiv.org/pdf/1905.08772.pdf)
 
-        Example
-        >>> clf.sg("chicken", "food")
-        >>> clf.sg("roast chicken", "food")
-        >>> clf.sg("chicken", "sports")
+        Examples:
+
+            >>> clf.sg("chicken", "food")
+            >>> clf.sg("roast chicken", "food")
+            >>> clf.sg("chicken", "sports")
 
         :param ngram: the word or word n-gram
         :type ngram: str
@@ -2401,10 +2501,11 @@ class SS3:
         (sn function is defined in Section 3.2.2 of the original paper:
          https://arxiv.org/pdf/1905.08772.pdf)
 
-        Example
-        >>> clf.sn("chicken", "food")
-        >>> clf.sn("roast chicken", "food")
-        >>> clf.sn("chicken", "sports")
+        Examples:
+
+            >>> clf.sn("chicken", "food")
+            >>> clf.sn("roast chicken", "food")
+            >>> clf.sn("chicken", "sports")
 
         :param ngram: the word or word n-gram
         :type ngram: str
@@ -2458,6 +2559,10 @@ def kmean_multilabel_size(res):
     clust = {"neg": [], "pos": []}  # clusters (2 clusters: "pos" and "neg")
     new_cent_neg = res[-1][1]
     new_cent_pos = res[0][1]
+
+    if new_cent_neg == new_cent_pos:
+        return 0
+
     while (cent["pos"] != new_cent_pos) or (cent["neg"] != new_cent_neg):
         cent["neg"], cent["pos"] = new_cent_neg, new_cent_pos
         clust["neg"], clust["pos"] = [], []
