@@ -11,9 +11,11 @@ from tqdm import tqdm
 from select import select
 from datetime import datetime
 
-from . import SS3, __version__
-from .util import RecursiveDefaultDict, Print, VERBOSITY
+from . import SS3, kmean_multilabel_size, __version__
+from .util import is_a_collection, membership_matrix, VERBOSITY
+from .util import Dataset, Print, RecursiveDefaultDict
 
+import numpy as np
 import webbrowser
 import argparse
 import socket
@@ -50,7 +52,7 @@ CONTET_TYPE = {
 }
 
 ENCODING = "utf-8"
-BASE_PATH = path.join(path.dirname(__file__), "resources/visual_classifier")
+BASE_PATH = path.join(path.dirname(__file__), "resources/live_test")
 
 
 def main():
@@ -59,6 +61,10 @@ def main():
 
     parser.add_argument('MODEL', help="the model name")
     parser.add_argument('-ph', '--path', help="the test set path")
+    parser.add_argument(
+        '-phl', '--path-labels', default=None,
+        help="the path to the labels (only for multilabel)"
+    )
     parser.add_argument(
         '-l', '--label', choices=["file", "folder"], default="folder",
         help="indicates where to read category labels from"
@@ -71,6 +77,7 @@ def main():
     )
     args = parser.parse_args()
 
+    Print.set_verbosity(VERBOSITY.VERBOSE)
     if args.quiet:
         Print.set_verbosity(VERBOSITY.QUIET)
 
@@ -89,7 +96,10 @@ def main():
 
     Server.set_model(clf)
     if args.path:
-        Server.set_testset_from_files(args.path, args.label == 'folder')
+        if args.path_labels is None:
+            Server.set_testset_from_files(args.path, args.label == 'folder')
+        else:
+            Server.set_testset_from_files_multilabel(args.path, args.path_labels)
 
     try:
         Server.serve(port=args.port, browser=False, quiet=args.quiet)
@@ -143,9 +153,14 @@ class Server:
 
     __x_test__ = None
     __test_path__ = None
+    __test_path_prev__ = None
+    __labels_path__ = None
+    __sep_doc__ = None
+    __sep_label__ = None
     __folder_label__ = None
     __preprocess__ = None
     __default_prep__ = None
+    __default_cat__ = None
 
     @staticmethod
     def __send_as_json__(sock, data):
@@ -249,8 +264,9 @@ class Server:
             "version": __version__,
             "model_name": clf.get_name(),
             "hps": clf.get_hyperparameters(),
-            "categories": clf.get_categories() + ["[unknown]"],
-            "docs": Server.__docs__
+            "categories": clf.get_categories(all=True) + ["[unknown]"],
+            "docs": Server.__docs__,
+            "def_cat": Server.__default_cat__
         })
         Print.info("sending classifier info...")
 
@@ -258,16 +274,7 @@ class Server:
     def __do_get_doc__(sock, file):
         """Serve the 'get_doc' message."""
         doc = ""
-        if ":line:" in file:
-            file, line_n = file.split(":line:")
-            line_n = int(line_n)
-            with open(file, 'r', encoding=ENCODING) as fdoc:
-                line = 0
-                doc = fdoc.readline()
-                while line < line_n:
-                    doc = fdoc.readline()
-                    line += 1
-        elif ":x_test:" in file:
+        if ":x_test:" in file:
             if Server.__x_test__ is not None:
                 idoc = int(file.split(":x_test:")[1])
                 doc = Server.__x_test__[idoc]
@@ -283,66 +290,69 @@ class Server:
         """Clear server's test documents."""
         Server.__docs__ = RecursiveDefaultDict()
         Server.__test_path__ = None
+        Server.__test_path_prev__ = None
         Server.__folder_label__ = None
         Server.__x_test__ = None
+        Server.__labels_path__ = None
+        Server.__sep_doc__ = None
+        Server.__sep_label__ = None
 
     @staticmethod
     def __load_testset_from_files__():
         """Load the test set files to visualize from ``test_path``."""
         Print.info("reading files...")
-        classify = Server.__clf__.classify
-        unkwon_cat_i = len(Server.__clf__.get_categories())
-        if not Server.__folder_label__:
-            for file in listdir(Server.__test_path__):
-                file_path = path.join(Server.__test_path__, file)
-                if path.isfile(file_path):
-                    cat = path.splitext(file)[0]
 
-                    with open(file_path, "r", encoding=ENCODING) as fcat:
-                        Server.__docs__[cat]["clf_result"] = [
-                            r[0][0] if r[0][1] else unkwon_cat_i
-                            for r in map(
-                                classify,
-                                tqdm(fcat.readlines(),
-                                     desc=" Classifying '%s' docs" % cat,
-                                     disable=Print.is_quiet())
-                            )
-                        ]
-                        n_docs = len(Server.__docs__[cat]["clf_result"])
-                        Server.__docs__[cat]["file"] = [
-                            "doc_%d" % line
-                            for line in range(n_docs)
-                        ]
-                        Server.__docs__[cat]["path"] = [
-                            "%s:line:%d" % (file_path, line)
-                            for line in range(n_docs)
-                        ]
+        if Server.__labels_path__:
+            docs_path = Server.__test_path__
+            labels_path = Server.__labels_path__
+            sep_label = Server.__sep_label__
+            x_test, y_test = Dataset.load_from_files_multilabel(docs_path, labels_path,
+                                                                sep_label, Server.__sep_doc__)
+            Server.set_testset(x_test, y_test)
+            if path.isdir(docs_path):
+                docs = Server.__docs__['']
+                Server.__x_test__ = None
+                sep_label = sep_label or r'\s+'  # default separator
+                with open(labels_path, "r", encoding=ENCODING) as flabels:
+                    doc_raw_names = [re.split(sep_label, l.rstrip())[0]
+                                     for l in flabels.read().split('\n')]
+                    doc_i = 0
+                    for doc_name in doc_raw_names:
+                        doc_name += ".txt"
+                        if doc_i == 0 or docs["file"][doc_i - 1] != doc_name:
+                            docs["file"][doc_i] = doc_name
+                            docs["path"][doc_i] = path.join(docs_path, doc_name)
+                            doc_i += 1
         else:
-            for cat in listdir(Server.__test_path__):
-                cat_path = path.join(Server.__test_path__, cat)
+            classify = Server.__clf__.classify
+            unkwon_cat_i = len(Server.__clf__.get_categories())
+            if not Server.__folder_label__:
+                x_test, y_test = Dataset.load_from_files(Server.__test_path__, False,
+                                                         sep_doc=Server.__sep_doc__)
+                Server.set_testset(x_test, y_test)
+            else:
+                for cat in listdir(Server.__test_path__):
+                    cat_path = path.join(Server.__test_path__, cat)
+                    if not path.isfile(cat_path):
+                        Server.__docs__[cat]["path"] = []
+                        Server.__docs__[cat]["file"] = []
+                        Server.__docs__[cat]["clf_result"] = []
+                        for file in tqdm(sorted(listdir(cat_path)),
+                                         desc=" Classifying '%s' docs" % cat,
+                                         disable=Print.is_quiet()):
+                            file_path = path.join(cat_path, file)
+                            if path.isfile(file_path):
+                                Server.__docs__[cat]["path"].append(file_path)
+                                Server.__docs__[cat]["file"].append(file)
+                                with open(
+                                    file_path, "r", encoding=ENCODING
+                                ) as fdoc:
+                                    r = classify(fdoc.read(), prep_func=Server.__preprocess__)
+                                    Server.__docs__[cat]["clf_result"].append(
+                                        r[0][0] if r[0][1] else unkwon_cat_i
+                                    )
 
-                if not path.isfile(cat_path):
-
-                    Server.__docs__[cat]["path"] = []
-                    Server.__docs__[cat]["file"] = []
-                    Server.__docs__[cat]["clf_result"] = []
-
-                    for file in tqdm(sorted(listdir(cat_path)),
-                                     desc=" Classifying '%s' docs" % cat,
-                                     disable=Print.is_quiet()):
-                        file_path = path.join(cat_path, file)
-                        if path.isfile(file_path):
-                            Server.__docs__[cat]["path"].append(file_path)
-                            Server.__docs__[cat]["file"].append(file)
-                            with open(
-                                file_path, "r", encoding=ENCODING
-                            ) as fdoc:
-                                r = classify(fdoc.read(), prep_func=Server.__preprocess__)
-                                Server.__docs__[cat]["clf_result"].append(
-                                    r[0][0] if r[0][1] else unkwon_cat_i
-                                )
-
-        Print.info("%d categories found" % len(Server.__docs__))
+            Print.info("%d categories found" % len(Server.__docs__))
         return len(Server.__docs__) > 0
 
     @staticmethod
@@ -367,7 +377,7 @@ class Server:
         Server.__clear_testset__()
 
     @staticmethod
-    def set_testset(x_test, y_test):
+    def set_testset(x_test, y_test=None, def_cat=None):
         """
          Assign the test set to visualize.
 
@@ -375,34 +385,75 @@ class Server:
         :type x_test: list (of str)
         :param y_label: the list of category labels
         :type y_label: list (of str)
+        :param def_cat: default category to be assigned when SS3 is not
+                        able to classify a document. Options are
+                        "most-probable", "unknown" or a given category name.
+                        (default: "most-probable", or "unknown" for
+                         multi-label classification)
+        :type def_cat: str
+        :raises: ValueError
         """
+        clf = Server.__clf__
+
         Server.__clear_testset__()
         Server.__x_test__ = x_test
+        Server.__default_cat__ = clf.__get_def_cat__(def_cat)
 
-        classify = Server.__clf__.classify
+        classify = clf.classify
+        docs = Server.__docs__
         unkwon_cat_i = len(Server.__clf__.get_categories())
+        no_y_test = y_test is None
+
+        multilabel = not no_y_test and len(y_test) and is_a_collection(y_test[0])
+        y_test_labels = y_test
+
+        if no_y_test or multilabel:
+            y_test = [""]
+
         for cat in set(y_test):
-            Server.__docs__[cat]["path"] = []
-            Server.__docs__[cat]["file"] = []
-            Server.__docs__[cat]["clf_result"] = []
+            docs[cat]["path"] = []
+            docs[cat]["file"] = []
+            docs[cat]["clf_result"] = []
 
         for idoc, doc in enumerate(x_test):
-            cat = y_test[idoc]
+            cat = y_test[0] if no_y_test or multilabel else y_test[idoc]
             doc_name = "doc_%d" % idoc
 
-            Server.__docs__[cat]["file"].append(doc_name)
-            Server.__docs__[cat]["path"].append(":x_test:%d" % idoc)
+            docs[cat]["file"].append(doc_name)
+            docs[cat]["path"].append(":x_test:%d" % idoc)
 
-            r = classify(doc, prep_func=Server.__preprocess__)
-            Server.__docs__[cat]["clf_result"].append(
-                r[0][0] if r[0][1] else unkwon_cat_i
-            )
+            if multilabel:
+                docs[cat]["clf_result"].append(
+                    classify(doc, prep_func=Server.__preprocess__)
+                )
+            elif not no_y_test:
+                res = classify(doc, prep_func=Server.__preprocess__)
+                docs[cat]["clf_result"].append(
+                    res[0][0] if res[0][1] else unkwon_cat_i
+                )
 
-        Print.info("%d categories found" % len(Server.__docs__))
-        return len(Server.__docs__) > 0
+        if multilabel:
+            y_pred = [[ci for ci, _ in r[:kmean_multilabel_size(r)]]
+                      for r in docs[y_test[0]]["clf_result"]]
+            if Server.__default_cat__ is not None:
+                y_pred = [labels if labels else [clf.get_category_index(Server.__default_cat__)]
+                          for labels in y_pred]
+            t = membership_matrix(clf, y_test_labels).todense()
+            p = membership_matrix(clf, y_pred, labels=False).todense()
+            np.seterr(divide='ignore', invalid='ignore')
+            accuracy = (t & p).sum(axis=1) / (t | p).sum(axis=1)
+            accuracy[np.isnan(accuracy)] = 1
+            docs[y_test[0]]["true_labels"] = y_test_labels
+            docs[y_test[0]]["labels_recall"] = accuracy.reshape(-1).tolist()[0]
+            docs[y_test[0]]["clf_result"] = [r[0][0] if r[0][1] else unkwon_cat_i
+                                             for r in docs[cat]["clf_result"]]
+
+        if not no_y_test and not multilabel:
+            Print.info("%d categories found" % len(docs))
+        return len(docs) > 0
 
     @staticmethod
-    def set_testset_from_files(test_path, folder_label=True):
+    def set_testset_from_files(test_path, folder_label=True, sep_doc='\n'):
         """
         Load the test set files to visualize from ``test_path``.
 
@@ -416,8 +467,10 @@ class Server:
         :rtype: bool
         """
         Server.__clear_testset__()
+        Server.__test_path_prev__ = Server.__test_path__
         Server.__test_path__ = test_path
         Server.__folder_label__ = folder_label
+        Server.__sep_doc__ = sep_doc
 
         docs = 0
         if not Server.__folder_label__:
@@ -436,6 +489,43 @@ class Server:
 
         return docs > 0
 
+    @staticmethod
+    def set_testset_from_files_multilabel(docs_path, labels_path, sep_label=None, sep_doc='\n'):
+        r"""
+        Multilabel version of the ``Live_Test.set_testset_from_files()`` function.
+
+        Load test documents and category labels from disk to visualize in the Live Test tool.
+
+        :param docs_path: the file or the folder containing the test documents.
+        :type docs_path: str
+        :param labels_path: the file containing the labels for each document.
+                            * if ``docs_path`` is a file, then the ``labels_path`` file
+                            should contain a line with the corresponding list of category
+                            labels for each document in ``docs_path``.
+                            * if ``docs_path`` is a folder containing the documents, then
+                            the ``labels_path`` file should contain a line for each document and
+                            category label. Each line should have the following format:
+                            ``document_name<sep_label>label``.
+
+        :type labels_path: str
+        :param sep_label: the separator/delimiter used to separate either each label (if
+                          ``docs_path`` is a file) or the document name from its category
+                          (if ``docs_path`` is a folder).
+                          (default: ``';'`` when ``docs_path`` is a file, the ``'\s+'`` regular
+                          expression otherwise).
+        :type sep_label: str
+        :param sep_doc: the separator/delimiter used to separate each document
+                        when loading training/test documents from single file. Valid
+                        only when ``folder_label=False``. (default: ``\n'``)
+        :type sep_doc: str
+        """
+        Server.__clear_testset__()
+        Server.__test_path_prev__ = Server.__test_path__
+        Server.__test_path__ = docs_path
+        Server.__labels_path__ = labels_path
+        Server.__sep_doc__ = sep_doc
+        Server.__sep_label__ = sep_label
+
     @staticmethod
     def start_listening(port=0):
         """
@@ -470,7 +560,7 @@ class Server:
     @staticmethod
     def serve(
         clf=None, x_test=None, y_test=None, port=0, browser=True,
-        quiet=True, prep=True, prep_func=None
+        quiet=True, prep=True, prep_func=None, def_cat=None
     ):
         """
         Wait for classification requests and serve them.
@@ -497,8 +587,16 @@ class Server:
                   If not given, the default preprocessing function will
                   be used
         :type prep_func: function
+        :param def_cat: default category to be assigned when SS3 is not
+                        able to classify a document. Options are
+                        "most-probable", "unknown" or a given category name.
+                        (default: "most-probable", or "unknown" for
+                         multi-label classification)
+        :type def_cat: str
+        :raises: ValueError
         """
-        Server.__clf__ = clf or Server.__clf__
+        clf = clf or Server.__clf__
+        Server.__clf__ = clf
         Server.__preprocess__ = prep_func
         Server.__default_prep__ = prep
 
@@ -510,12 +608,12 @@ class Server:
             Server.start_listening(port)
 
         if x_test is not None:
-            if y_test is not None and len(y_test) == len(x_test):
-                Server.set_testset(x_test, y_test)
+            if y_test is None or len(y_test) == len(x_test):
+                Server.set_testset(x_test, y_test, def_cat)
             else:
                 Print.error("y_test must have the same length as x_test")
                 return
-        elif Server.__test_path__ and not Server.__docs__:
+        elif Server.__test_path__ and Server.__test_path_prev__ != Server.__test_path__:
             Server.__load_testset_from_files__()
 
         server_socket = Server.__server_socket__
