@@ -2,7 +2,7 @@
 """Tests for pyss3."""
 from os import path
 from shutil import rmtree
-from pyss3.util import Dataset
+from pyss3.util import Dataset, membership_matrix
 from pyss3 import \
     SS3, STR_NORM_GV_XAI, STR_NORM_GV, STR_GV, \
     STR_XAI, STR_VANILLA, STR_MOST_PROBABLE, \
@@ -16,8 +16,10 @@ import pytest
 pyss3.set_verbosity(VERBOSITY.QUIET)
 
 DATASET_FOLDER = "dataset"
+DATASET_MULTILABEL_FOLDER = "dataset_ml"
 
 dataset_path = path.join(path.abspath(path.dirname(__file__)), DATASET_FOLDER)
+dataset_multilabel_path = path.join(path.abspath(path.dirname(__file__)), DATASET_MULTILABEL_FOLDER)
 
 x_train, y_train = Dataset.load_from_files(dataset_path, folder_label=False)
 x_test = [
@@ -91,8 +93,8 @@ def perform_tests_with(clf, cv_test, stopwords=True):
     most_prob_cat = clf.get_most_probable_category()
     most_prob_cat_idx = clf.__get_most_probable_category__()
 
-    # category names case-insensitiveness check
-    assert clf.get_category_index("SpOrTs") == clf.get_category_index("sports")
+    # category names case-sensitiveness check
+    assert clf.get_category_index("SpOrTs") == -1
 
     # predict
     y_pred = clf.predict(x_test)
@@ -113,9 +115,9 @@ def perform_tests_with(clf, cv_test, stopwords=True):
     y_pred = clf.predict([doc_unknown], def_cat=STR_UNKNOWN)
     assert y_pred[0] == STR_UNKNOWN_CATEGORY
     y_pred = clf.predict([doc_unknown], multilabel=True)
-    assert y_pred[0] == [most_prob_cat]
-    y_pred = clf.predict([doc_unknown], def_cat=STR_UNKNOWN, multilabel=True, labels=False)
-    assert y_pred[0] == [IDX_UNKNOWN_CATEGORY]
+    assert y_pred[0] == []
+    y_pred = clf.predict([doc_unknown], def_cat=STR_MOST_PROBABLE, multilabel=True, labels=False)
+    assert y_pred[0] == [most_prob_cat_idx]
 
     y_pred = clf.predict([doc_unknown], def_cat=STR_MOST_PROBABLE)
     assert y_pred[0] == most_prob_cat
@@ -173,16 +175,17 @@ def perform_tests_with(clf, cv_test, stopwords=True):
     assert len(multilabel_labels) == len(r)
     assert r[0] in multilabel_idxs and r[1] in multilabel_idxs
 
-    assert clf.classify_multilabel('') == [most_prob_cat]
-    assert clf.classify_multilabel('', def_cat=STR_UNKNOWN) == [pyss3.STR_UNKNOWN_CATEGORY]
+    assert clf.classify_multilabel('') == []
+    assert clf.classify_multilabel('', def_cat=STR_MOST_PROBABLE) == [most_prob_cat]
     assert clf.classify_multilabel('', def_cat=def_cat) == [def_cat]
 
-    assert clf.classify_multilabel(doc_unknown) == [most_prob_cat]
-    assert clf.classify_multilabel(doc_unknown, def_cat=STR_UNKNOWN) == [pyss3.STR_UNKNOWN_CATEGORY]
+    assert clf.classify_multilabel(doc_unknown) == []
+    assert clf.classify_multilabel(doc_unknown, def_cat=STR_MOST_PROBABLE) == [most_prob_cat]
     assert clf.classify_multilabel(doc_unknown, def_cat=def_cat) == [def_cat]
 
-    assert clf.classify_multilabel(doc_unknown, labels=False) == [most_prob_cat_idx]
-    assert clf.classify_multilabel(doc_unknown, def_cat=STR_UNKNOWN, labels=False) == [-1]
+    assert clf.classify_multilabel(doc_unknown, labels=False) == []
+    assert clf.classify_multilabel(doc_unknown, def_cat=STR_MOST_PROBABLE,
+                                   labels=False) == [most_prob_cat_idx]
     assert clf.classify_multilabel(doc_unknown, def_cat=def_cat, labels=False) == [def_cat_idx]
 
     # "learn an doc_unknown and a new category" case
@@ -247,14 +250,51 @@ def test_pyss3_functions():
     r = [(6, 8.1), (7, 5.6), (2, 5.5), (4, 1.5),
          (5, 1.3), (3, 1.2), (0, 1.1), (1, 0.4)]
     assert pyss3.kmean_multilabel_size(r) == 3
-    assert pyss3.kmean_multilabel_size([(0, 0), (1, 0)]) == 2
+    assert pyss3.kmean_multilabel_size([(0, 0), (1, 0)]) == 0
+    assert pyss3.kmean_multilabel_size([(0, 10), (1, 10)]) == 0
 
     with pytest.raises(IndexError):
         pyss3.mad([], 0)
 
 
+def test_multilabel():
+    """Test multilabel support."""
+    x_train, y_train = Dataset.load_from_files_multilabel(
+        path.join(dataset_multilabel_path, "train/docs.txt"),
+        path.join(dataset_multilabel_path, "train/labels.txt"),
+        sep_label=",",
+        sep_doc="\n>>>>>\n"
+    )
+
+    clf = SS3()
+
+    with pytest.raises(ValueError):
+        membership_matrix(clf, [])
+
+    clf.fit(x_train, y_train)
+
+    assert sorted(clf.get_categories()) == ['insult', 'obscene', 'severe_toxic', 'toxic']
+    assert clf.classify_multilabel("this is a unknown document!") == []
+
+    y_pred = [[], ['toxic'], ['severe_toxic'], ['obscene'], ['insult'], ['toxic', 'insult']]
+
+    y_pred_memmatrix = membership_matrix(clf, y_pred).todense().tolist()
+    assert y_pred_memmatrix == [[0, 0, 0, 0],  # []
+                                [1, 0, 0, 0],  # ['toxic']
+                                [0, 1, 0, 0],  # ['severe_toxic']
+                                [0, 0, 1, 0],  # ['obscene']
+                                [0, 0, 0, 1],  # ['insult']
+                                [1, 0, 0, 1]]  # ['toxic', 'insult']
+
+    y_pred_memmatrix = membership_matrix(clf, y_pred + [["xxx"]]).todense().tolist()
+    assert y_pred_memmatrix[-1] == [0, 0, 0, 0]
+
+
 def test_pyss3_ss3(mockers):
     """Test SS3."""
+    with pytest.raises(ValueError):
+        clf = SS3("hyperparameter")
+
     clf = SS3(
         s=.45, l=.5, p=1, a=0,
         cv_m=STR_NORM_GV_XAI, sn_m=STR_XAI
@@ -274,6 +314,10 @@ def test_pyss3_ss3(mockers):
         clf.predict(x_test)
     with pytest.raises(pyss3.EmptyModelError):
         clf.predict_proba(x_test)
+    with pytest.raises(ValueError):
+        clf.train(x_train, [])
+    with pytest.raises(ValueError):
+        clf.train([], [])
 
     # train and predict/classify tests (model: terms are single words)
     # cv_m=STR_NORM_GV_XAI, sn_m=STR_XAI
